@@ -92,7 +92,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
             use checkout_core::checkout::Checkout;
             use checkout_core::payment::Payment;
             use checkout_core::payment::PaymentProcessor;
-            use checkout_core::payment::handlers::process_handler;
+            use checkout_core::payment::handlers::{process_handler, step_handler};
             use checkout_core::customer::Contact;
             use checkout_core::customer::Address;
             use checkout_core::shipping::FulfillmentTypeSelection;
@@ -158,9 +158,12 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                                                     if !ctx.start_transaction().await.is_ok() {
                                                         continue;
                                                     }
-                                                    let processing_result = process_handler(&internal_ctx, &mut ctx, message.0).await;
+                                                    let processing_result = process_handler(&internal_ctx, &mut ctx, &message.0).await;
                                                     if !processing_result.is_ok() {
-                                                        ctx.abort_transaction().await;
+                                                        if !ctx.abort_transaction().await.is_ok() {
+                                                            // just log the failure to abort tx
+                                                            eprintln!("failed to commit tx");
+                                                        }
                                                         continue;
                                                     }
                                                     if !ctx.commit_transaction().await.is_ok() {
@@ -175,10 +178,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                                     }
                                 }
                             }
-                            Err(e) => {
-                                eprintln!("Failed to pull PubSub messages: {}", e);
-                                println!("Failed to pull PubSub messages: {}", e);
-                            },
+                            Err(e) => eprintln!("Failed to pull PubSub messages: {}", e)
                         }
                     }
                 });
@@ -468,6 +468,57 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                 }
             }
+
+            // payment related endpoints
+            // struct CorrelationIdParam {
+            //     id: <super::#context_type_name<'_> as PaymentProcessor>::CorrelationID
+            // }
+
+            #[endpoint {
+                method = PATCH,
+                path = "/payment/{id}/step_payment",
+            }]
+            pub async fn step_payment_endpoint(
+                req_ctx: Arc<RequestContext<ServerContext>>,
+                path: Path<IdParam>,
+                body: TypedBody<<super::#context_type_name<'_> as PaymentProcessor>::ProcessArgs>,
+            ) -> Result<HttpResponseOk<Payment<
+                <super::#context_type_name<'_> as PaymentProcessor>::CorrelationID,
+                <super::#context_type_name<'_> as PaymentProcessor>::PrivateData,
+                <super::#context_type_name<'_> as PaymentProcessor>::PublicData,
+                <super::#context_type_name<'_> as PaymentProcessor>::ProcessArgs
+            >>, HttpError> {
+                let path_params = path.into_inner();
+                let args = body.into_inner();
+                let internal_ctx = req_ctx
+                    .context()
+                    .internal_context
+                    .clone();
+
+                let mut ctx = req_ctx
+                    .context()
+                    .context_provider
+                    .new_context()
+                    .await
+                    .unwrap();
+
+                ctx.start_transaction().await.unwrap();
+                match step_handler(&internal_ctx, &mut ctx, &path_params.id, args).await {
+                    Ok(co) => {
+                        ctx.commit_transaction().await.unwrap();
+                        Ok(HttpResponseOk(co))
+                    }
+                    Err(err) => {
+                        ctx.abort_transaction().await.unwrap();
+                        Err(HttpError {
+                            status_code: StatusCode::BAD_REQUEST,
+                            error_code: Some(err.code.clone()),
+                            external_message: err.message.clone(),
+                            internal_message: err.message.clone(),
+                        })
+                    }
+                }
+            }
         }
 
         #[async_trait]
@@ -484,6 +535,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 api.register(checkout_server::update_shipping_address_endpoint).unwrap();
                 api.register(checkout_server::update_fulfillment_type_endpoint).unwrap();
                 api.register(checkout_server::initiate_payment_endpoint).unwrap();
+                api.register(checkout_server::step_payment_endpoint).unwrap();
 
                 api
             }
